#include "../OsaRtl.h"
 
// Headers that cannot be included in C++ code even with macro fixes
// hence we need to provide C functions here for which we give access to the C++ parts
// using our header (OsaLinuxRtl.h), that is written so that it compiles both in C and C++
// (with extern "C" linkage in C++, of course)
#include <linux/malloc.h>
#include <asm/io.h>
#include <linux/interrupt.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0) 
#include <linux/poll.h>
#else
#endif

static spinlock_t ms_accessInterlocked;

static OsaBool32  ms_bUse1601TypeTimeStamp = OSA_TRUE;

/******************************************************************************
******************************************************************************/
void OsaInitLinuxRtl(void)
{
  ms_accessInterlocked = ms_accessInterlocked;
  spin_lock_init(&ms_accessInterlocked);  

  ms_bUse1601TypeTimeStamp = OSA_TRUE;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
/******************************************************************************
******************************************************************************/
int osalinux_wait_event_timeout(
  OsaUint32 *pEventSet, 
  struct wait_queue **ppWaitQueue, 
  OsaUint32 nWaitMaxMs
)
{
  int rc = 0;
  struct wait_queue sWaitQueue; 
  sWaitQueue.task = current;       
  add_wait_queue(ppWaitQueue, &sWaitQueue); 
  
  OSA_DPRINTF(("OsaLinux: osalinux_wait_event_timeout \n"));
  
  for (;;) {
    current->state = TASK_INTERRUPTIBLE;
    if (*pEventSet) break;
    if (!signal_pending(current)) {
      if (!schedule_timeout(nWaitMaxMs / (1000 / HZ))) {
        rc = OSA_RES_TIMEOUT;  // Timeout
        break;
      }
      continue;
    }
    rc = OSA_RES_FAILED;
    break;
  }
  current->state = TASK_RUNNING;
  remove_wait_queue(ppWaitQueue, &sWaitQueue);
  return rc;
}

#else
/******************************************************************************
******************************************************************************/
int osalinux_wait_event_timeout(
  OsaUint32 *pEventSet, 
  wait_queue_head_t *pWaitQueueHead,             
  OsaUint32 nWaitMaxMs
)
{
  int rc = 0;
  wait_queue_t sWaitQueue;                       
  sWaitQueue.task = current;                      
  add_wait_queue(pWaitQueueHead, &sWaitQueue);   

  OSA_DPRINTF(("OsaLinux: osalinux_wait_event_timeout \n"));
  
  for (;;) {
    current->state = TASK_INTERRUPTIBLE;
    if (*pEventSet) break;
      if (!signal_pending(current)) {
       if (!schedule_timeout(nWaitMaxMs / (1000 / HZ))) {
         rc = OSA_RES_TIMEOUT;  // Timeout
         break;
      }
      continue;
    }
    rc = OSA_RES_FAILED;
    break;
  }
  current->state = TASK_RUNNING;

  remove_wait_queue(pWaitQueueHead, &sWaitQueue);	
  
	return rc;
}
#endif

/******************************************************************************
******************************************************************************/
void *osalinux_kmalloc(size_t size, int bAtomic)
{
#if 0 // kmalloc() can only allocate 128k on intel platforms. We need more.
  if (bAtomic) {
    return kmalloc(size, GFP_ATOMIC); 
  } else {
    return kmalloc(size, GFP_KERNEL); 
  }
#else
  return vmalloc(size);
#endif
}


/******************************************************************************
******************************************************************************/
void osalinux_kfree(void *ptr)
{
#if 0 // kmalloc() can only allocate 128k on intel platforms. We need more.
  kfree(ptr);
#else
  return vfree(ptr);
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
/******************************************************************************
******************************************************************************/
void osalinux_wake_up(struct wait_queue **wait_queue)
{
  wake_up(wait_queue);
}


/******************************************************************************
******************************************************************************/
void osalinux_poll_wait(struct file * filp, struct wait_queue ** wait_address, struct poll_table_struct *p)
{
  poll_wait(filp, wait_address, p);
}

#else
/******************************************************************************
******************************************************************************/
void osalinux_wake_up(wait_queue_head_t *wait_queue_head)  
{
  wake_up(wait_queue_head);
}


/******************************************************************************
******************************************************************************/
void osalinux_poll_wait(struct file * filp, wait_queue_head_t * wait_address, struct poll_table_struct *p)	
{
  poll_wait(filp, wait_address, p);
}
#endif


/******************************************************************************
******************************************************************************/
int osalinux_request_irq(unsigned int irq,  void (*handler)(int, void *, struct pt_regs *), void *dev_id)
{
  return request_irq(irq, handler, SA_INTERRUPT | SA_SHIRQ, "OSA_Device", dev_id); 
}


/******************************************************************************
******************************************************************************/
void osalinux_free_irq(unsigned int irq, void *dev_id)
{
  free_irq(irq, dev_id);
}


/******************************************************************************
******************************************************************************/
void *osalinux_ioremap(unsigned long offset, unsigned long size)
{
  return ioremap(offset, size);
}


/******************************************************************************
******************************************************************************/
void osalinux_iounmap(void *addr)
{
  iounmap(addr);
}

/******************************************************************************
******************************************************************************/
wchar_t *wcscpy (wchar_t *dest, const wchar_t *source)
{
  wchar_t *d = dest;

  while (*source) {
    *(dest++) = *(source++);
  }

  *dest = 0;

  return d;
}


/******************************************************************************
******************************************************************************/
size_t wcslen(const wchar_t *psz)
{
  size_t r = 0;

  while (*(psz++)) {
    //OSA_DPRINTF(("OsaLinux: %c", (char) *str));
    r++;
  }

  //OSA_DPRINTF(("\n"));
  return r;
}

#if 0
/******************************************************************************
******************************************************************************/
wchar_t *wcscat (wchar_t *dest, const wchar_t *source)
{

  while (*dest)
    dest++;

  while (*source) {
    *(dest++) = *(source++);
  }

  *dest = 0x0;

  return dest;
}
#endif


/******************************************************************************
******************************************************************************/
wchar_t *wcsncpy(wchar_t *dest, const wchar_t *source, size_t count)
{
  wchar_t *d = dest;

  while (*source && count--) {
    *(dest++) = *(source++);
  }

  *dest = 0;

  return d;
}


/******************************************************************************

  @comm We rely on the fact that all 7 bit ASCII chars have the same numeric
  representation in the Unicode char set. We only handle conversion of
  these "code 127 or less" type Unicode characters, stored in wchar_t strings,
  to ASCII stored in char strings.

******************************************************************************/
char *wcsztosz(char *strDest, const wchar_t *strSource)
{
  while (*strSource) {
    *(strDest++) = (char) *(strSource++);
  }

  *strDest = '\0';

  return strDest; // Returning a pointer to the end allows easy for concatenation
}


/******************************************************************************
******************************************************************************/
wchar_t towupper (wchar_t c)
{
  // Only handles US English
  if (c >= L'a' && c <= L'z') {
    return c - (L'a' - L'A');
  }
  return c;
}


/******************************************************************************
******************************************************************************/
char *_strrev(char *szBuf)
{
  char chr;
  char *pszStart = szBuf;
  char *pszEnd = szBuf + strlen(szBuf) - 1;
  
  while (pszStart < pszEnd) {
    chr = *pszStart;
    *pszStart = *pszEnd;
    *pszEnd = chr;
    
    pszEnd--; pszStart++;
  }
  
  return szBuf;
}


/*******************************************************************************
*******************************************************************************/
OsaUint32 OsaIrpGetIoCode(IN OsaIrpP pIrp)
{
  return pIrp->nIoCode;
}


/******************************************************************************
******************************************************************************/
OsaVoidP OsaIrpGetInputBuffer(IN OsaIrpP pIrp)
{
  return pIrp->pInputBuffer;
}


/******************************************************************************
******************************************************************************/
OsaVoidP OsaIrpGetOutputBuffer(IN OsaIrpP pIrp)
{
  return pIrp->pOutputBuffer;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetOwnerId(IN OsaIrpP pIrp)
{
  return pIrp->nOwnerId;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetInputBufferSize(IN OsaIrpP pIrp)
{
  return pIrp->cbInput;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetOutputBufferCapacity(IN OsaIrpP pIrp)
{
  return pIrp->cbOutputCapacity;
}


/******************************************************************************
******************************************************************************/
OsaResultE  OsaIrpComplete(IN OsaIrpP pIrp, IN OsaResultE eResult, IN OsaUint32 cbOutput)
{
  pIrp->result = eResult;
  pIrp->cbOutput = cbOutput;
  return eResult;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaReadReg32(OsaUint32 pRegister)
{
  return readl(pRegister);
}


/******************************************************************************
******************************************************************************/
OsaUint16 OsaReadReg16(OsaUint32 pRegister)
{
  return readw(pRegister);
}


/******************************************************************************
******************************************************************************/
OsaUint8 OsaReadReg8 (OsaUint32 pRegister)
{
  return readb(pRegister);
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg32(OsaUint32 pRegister, OsaUint32 nValue)
{
  writel(nValue, pRegister);
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg16(OsaUint32 pRegister, OsaUint16 nValue)
{
  writew(nValue, pRegister);
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg8 (OsaUint32 pRegister, OsaUint8  nValue)
{
  writeb(nValue, pRegister);
}


/******************************************************************************
******************************************************************************/
void OsaReadRegBuffer8  (OsaUint32 pRegister, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  OsaUint32 coReads;                                                
  for (coReads = 0; coReads < coCount; coReads++, pBuffer++) { 
    *pBuffer = readb(pRegister);
  }                                                               
}


/******************************************************************************
******************************************************************************/
void OsaWriteRegBuffer8 (OsaUint32 pRegister, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  OsaUint32 coWrites;                                          
  for (coWrites = 0; coWrites < coCount; coWrites++, pBuffer++) {
    writeb(*pBuffer, pRegister);
  }                                                             
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaReadPort32(OsaUint32 pPort)
{
  return inl(pPort);
}


/******************************************************************************
******************************************************************************/
OsaUint16 OsaReadPort16(OsaUint32 pPort)
{
  return inw(pPort);
}


/******************************************************************************
******************************************************************************/
OsaUint8 OsaReadPort8 (OsaUint32 pPort)
{
  return inb(pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort32(OsaUint32 pPort, OsaUint32 nValue)
{
  outl(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort16(OsaUint32 pPort, OsaUint16 nValue)
{
  outw(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort8 (OsaUint32 pPort, OsaUint8  nValue)
{
  outb(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaReadPortBuffer8 (OsaUint32 pPort, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  insb(pPort, pBuffer, coCount); 
}


/******************************************************************************
******************************************************************************/
void OsaWritePortBuffer8(OsaUint32 pPort, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  outsb(pPort, pBuffer, coCount); 
}


/******************************************************************************
  @devnote Note that the Osa Interlocked functions are only guaranteed to be
  interlocked in relation to eachother.
******************************************************************************/
OsaUint8 OsaInterlockedModifyRegBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits)
{
  OsaUint8 nValue;
  __u32 nSavedFlags;

  spin_lock_irqsave(&ms_accessInterlocked, nSavedFlags);
  nValue = readb(pRegister);

  // Clear the bits we modify
  nValue &= ~nBitsToModify;
  nValue |= nValueOfBits;

#if DBG
  // Make sure the value doesn't modify bits it shouldn't
  OSA_ASSERT((nValueOfBits & ~nBitsToModify) == 0);
#endif

  writeb(nValue, pRegister);

  spin_unlock_irqrestore(&ms_accessInterlocked, nSavedFlags);

  return nValue;
}


/******************************************************************************
  @devnote Note that the Osa Interlocked functions are only guaranteed to be
  interlocked in relation to eachother.
******************************************************************************/
OsaUint8 OsaInterlockedModifyPortBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits)
{
  OsaUint8 nValue;
  __u32 nSavedFlags;

  spin_lock_irqsave(&ms_accessInterlocked, nSavedFlags);
  nValue = inb(pRegister);

  // Clear the bits we modify
  nValue &= ~nBitsToModify;
  nValue |= nValueOfBits;

#if DBG
  // Make sure the value doesn't modify bits it shouldn't
  OSA_ASSERT((nValueOfBits & ~nBitsToModify) == 0);
#endif

  outb(nValue, pRegister);

  spin_unlock_irqrestore(&ms_accessInterlocked, nSavedFlags);

  return nValue;
}


/******************************************************************************
******************************************************************************/
void OsaInterlockedIncrement(IN OUT OsaInterlockedIntP pValue)
{
  atomic_inc(pValue);
}


/******************************************************************************
******************************************************************************/
OsaInt32 OsaInterlockedDecrement(IN OUT OsaInterlockedIntP pValue)
{
  atomic_dec(pValue);
  return pValue->counter;
  //return 0;
}


/******************************************************************************
    Suspends execution for a specified time (in microseconds).
******************************************************************************/
void OsaDelayThreadMicroSecs(
  OsaInt64 nMicroSecs     // @parm Time to suspend the execution of the current thread
)
{
  OSA_ASSERT(nMicroSecs >= 0);

  if (nMicroSecs <= 1000*MAX_UDELAY_MS) {
    udelay(nMicroSecs);  // Simply busy wait
  } else if (nMicroSecs <= 1000000/HZ) {
    // longer wait time bet less than a clock tick
    mdelay(((OsaUint32) nMicroSecs) / 1000);   // use mdelay for full msecs (udelay can overflow)
    udelay(((OsaUint32) nMicroSecs) % 1000);
  } else {
    // Long wait times > 1 clock tick, we can use the scheduler
    // Note the tricks to avoid 64 bit division and modulo, which gcc cannot handle without emitting __divdi and __moddi
    // which are defned in libgcc.a
    
    OsaUint32 highPart = (OsaUint32) (nMicroSecs >> 32);
    OsaUint32 lowPart  = (OsaUint32) (nMicroSecs & 0xFFFFFFFF);
                              
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(lowPart / (1000000L / HZ));

    while (highPart--) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(0xFFFFFFFF / (1000000L / HZ));
    }
  }

}


/******************************************************************************
  Suspends execution for a specified time (in microseconds).
  Can be used from all OSA_EXE_LEVELs, but must not be used for delays
  longer than 50 microsecs (i.e. "hardware"-type delays)
******************************************************************************/
void OsaStallExecutionMicroSecs(
  IN OsaUint32 nMicroSecs       // @parm Time to suspend the execution of the processor
)
{
  OSA_ASSERT(nMicroSecs < 50);    // Must be less than 50 because otherwise we will miss a clock tick.
  udelay(nMicroSecs);            // Simply busy wait
}


/******************************************************************************

  @devnote See Rubini pp. 396

******************************************************************************/
OsaExeLevel OsaGetCurrentExeLevel()
{
  if (in_interrupt()) {
    return OSA_EXE_LEVEL_SYSTEM;
  } else {
    return OSA_EXE_LEVEL_CONTROL;
  }
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaGetSystemTimeTickFrequency(void)
{
  // In Linux we only support one tick frequency (100ns ticks)
  return 10000000;
}

/******************************************************************************
******************************************************************************/
void OsaSetSystemTimeSource(OsaBool32 bUse1601TypeTimeStamp)
{
  // Not used in Linux
  ms_bUse1601TypeTimeStamp = bUse1601TypeTimeStamp;
}


/******************************************************************************
  Microsecond resolution under Linux 2.0 and greater on Intel.

  This function should return 100 nanosecond increments since absolute
  time since Jan 1, 1601 but now it does it since midnight only

******************************************************************************/
void OsaQuerySystemTime(OsaDateTime *pDateTime)
{
  struct timeval tv;
  do_gettimeofday (&tv);
  pDateTime->nUint64 = tv.tv_sec;
  pDateTime->nUint64 *= 10000000L;
  pDateTime->nUint64 += tv.tv_usec * 10;
}


/******************************************************************************
  100th of a second resolution (typically)
******************************************************************************/
OsaUint32 OsaQueryClockMs(void)
{
  if (HZ < 1000) {
    return jiffies * (1000/HZ);
  } else {
    return jiffies / (HZ/1000);
  }
}


/******************************************************************************
		Return bits (n) starting from bit position p, right adjusted.
  
		Bit position 0 is at the right end, bit 31 is at the left end
		-----------------------------------------------------------------
			3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1  
			1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
		-----------------------------------------------------------------
******************************************************************************/
OsaUint32 getbits(OsaUint32 x, int p, int n)
{
	if( (x == 1) && (n == 32) )	
	{
		return 1;
	}
	else if ( ( (x >> 31) == 1) && (n == 32) ) 
	{
		return x;
	}
	else 
	{
		return (x >> (p+1-n)) & ~(~0 << n);	  
	}					
}



/****************************************************************************** 
	Unsigned 64-bit divided by unsigned 32-bit 
	
	Utilize bitwise operator to perform division 
******************************************************************************/
OsaUint64 OsaInt64div32(OsaUint64 divident, OsaUint32 divider)
{
	int i = 1, k = 0, flag = 0;
	OsaDateTime tempDivident, tempDivider, temp, result, tempRes, med;

	temp.nUint64    = 0;
	med.nUint64		  = 0;
	result.nUint64	= 0;
	tempRes.nUint64 = 0;

	tempDivident.nUint64 = divident;
  tempDivider.nUint64  = divider;
	
	// check dividor 
	if( tempDivider.nUint64 == 1 ) 
	{			
		return tempDivident.nUint64;
	}
	else if( tempDivider.nUint64 == 0 ) 
	{
		return ( 1 / 0);
	}
	
	// check divident whether it is 64-bit or not
	if( tempDivident.sUint32.nHigh != 0)
		flag = 1;


	if( flag == 0 )	// 32-bit
	{
		result.nUint64 = tempDivident.sUint32.nLow / tempDivider.sUint32.nLow;
	}
	else						//	64-bit
	{
		int dividentHighMsbPos	= 0;
		int dividentHighBitSize = 0;
		int dividentLowMsbPos		= 31;		
		int dividerBitSize = 0;

		// get High section of MSB's position of dividend 
		while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1	)
		{
			if( flag == 1 ) 
				break;
		}
		dividentHighMsbPos = 31 - (i - 1) + 1;
		dividentHighBitSize = dividentHighMsbPos + 1;
	
		// get low section of MSB's position of divider
		while( (getbits(tempDivider.sUint32.nLow, 31, k++) & 1) != 1	);
		dividerBitSize = (31 - (k - 1) + 1) + 1;


		// when bit size of High is larger than the size of divider
		while( dividentHighBitSize > dividerBitSize )
		{			
			k = 0;

			temp.nUint64 = 0;		
			temp.sUint32.nHigh = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividerBitSize);

			if(temp.sUint32.nHigh < tempDivider.sUint32.nLow)
			{
				k++;
				temp.sUint32.nHigh = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividerBitSize + k);
			}

			dividentHighMsbPos--;
			dividentHighBitSize = dividentHighMsbPos + 1;
			
			// romove bits that already been used
			tempDivident.sUint32.nHigh &= ~(~0 << (dividentHighBitSize + 1 - dividerBitSize - k) );	 

			// get remainder
			temp.sUint32.nHigh -= tempDivider.sUint32.nLow;
			temp.sUint32.nHigh = ( temp.sUint32.nHigh << ( dividentHighBitSize + 1 - dividerBitSize - k ) ); 
			
			// add remainder to dividend 
			tempDivident.sUint32.nHigh |= (getbits(temp.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize));
	
			// take result
			result.sUint32.nHigh |= (1 << ( dividentHighBitSize + 1 - dividerBitSize - k ) );	  

			// get High section of MSB's position of dividend -- since dividentHighBitSize is changed 
			i = 1;
			while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )   
			{
				if( (i == 33) || (flag == 1) )
					break;
			}
			dividentHighMsbPos = 31 - (i - 1) + 1;
			dividentHighBitSize = dividentHighMsbPos + 1;
			
		}	// while

		tempRes.nUint64 = result.nUint64;


		// get High section of MSB's position of dividend 
		i = 1;
		while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )   
		{
			if( (i == 33) || (flag == 1) )
				break;
		}
			
		if ( ( i == 33 ) && ( flag != 1 ) )		// no High bit remained 
		{
			dividentHighMsbPos = -1;
			dividentHighBitSize = dividentHighMsbPos + 1;
		}
		else
		{
			dividentHighMsbPos = 31 - (i - 1) + 1;
			dividentHighBitSize = dividentHighMsbPos + 1;
		}


		// when bit size of High is less than or equal to the size of divider
		while( (dividentHighBitSize <= dividerBitSize) && (dividentHighBitSize > 0) )
		{
			temp.nUint64 = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize); 

			if( temp.nUint64 == tempDivider.sUint32.nLow )
			{
				dividentHighMsbPos = dividentHighMsbPos - dividerBitSize;
				dividentHighBitSize = dividentHighMsbPos + 1;

				tempDivident.sUint32.nHigh &= 0;
				result.sUint32.nHigh = 1;
			}
			else if( temp.nUint64 < tempDivider.sUint32.nLow )
			{
				k = 0;
				
				// get bits which will be subtracted by divider
				if( dividerBitSize == dividentHighBitSize )
				{
					temp.nUint64 <<= 1;
					med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize + 1));
				}
				else
				{
					temp.nUint64 = (temp.nUint64 << (dividerBitSize - dividentHighBitSize) );
					med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize));
				}

				med.nUint64 |= temp.nUint64;

				if(med.nUint64 < tempDivider.sUint32.nLow)
				{
					k++;

					temp.nUint64 <<= 1;
					med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize) + 1);
					med.nUint64 |= temp.nUint64;
				}

				dividentHighMsbPos--;
				dividentHighBitSize = dividentHighMsbPos + 1;
		
				// remove High bits that are used
				tempDivident.sUint32.nHigh &= 0;
				

				if( dividerBitSize == (dividentHighBitSize + 1) )	// +1 is because dividentnHighMsbPos--
				{
					tempDivident.sUint32.nLow &= ~(~0 << (31 - (dividerBitSize - ( dividentHighBitSize + 1)) ));
				}
				else
				{
					tempDivident.sUint32.nLow &= ~(~0 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k ) );
				}

				med.nUint64 -= tempDivider.sUint32.nLow;
					
				if( dividerBitSize == (dividentHighBitSize + 1) )
					med.nUint64 = ( med.nUint64 << (31 - (dividerBitSize - ( dividentHighBitSize + 1))) );
				else
					med.nUint64 = ( med.nUint64 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k ) );
	
				// add remainder to dividend 
				tempDivident.sUint32.nLow  |=	med.sUint32.nLow;
				tempDivident.sUint32.nHigh |= med.sUint32.nHigh;

				// take result
				if( dividerBitSize == (dividentHighBitSize + 1) )
					result.sUint32.nLow |= (1 << (31 - (dividerBitSize - ( dividentHighBitSize + 1))) );
				else
					result.sUint32.nLow |= (1 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k) );


				if(dividentHighBitSize <= 0)
					break;
			}
			else if( temp.nUint64 > tempDivider.sUint32.nLow )
			{
				dividentHighMsbPos--;
				dividentHighBitSize = dividentHighMsbPos + 1;

				// romove bits that already used
				tempDivident.sUint32.nHigh &= 0;

				temp.nUint64 -= tempDivider.sUint32.nLow;
				temp.nUint64 = ( temp.nUint64 << 32 );
	
				// add remainder to dividend 
				tempDivident.sUint32.nHigh |= (getbits(temp.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize));

				// take result
				result.sUint32.nHigh |= 1;
			}

			i = 1;
			while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )   
			{
				if( ( i == 33 ) || ( flag == 1 ) )
					break;
			}
			
			if ( ( i == 33 ) && ( flag != 1 ) )
			{
				dividentHighMsbPos = -1;
				dividentHighBitSize = dividentHighMsbPos + 1;
			}
			else
			{
				dividentHighMsbPos = 31 - (i - 1) + 1;
				dividentHighBitSize = dividentHighMsbPos + 1;
			}

		} // while divident bit size > divider bit size

		// 32-bit remainder
		temp.nUint64 = tempDivident.sUint32.nLow / tempDivider.sUint32.nLow;
		
		result.nUint64 |= tempRes.nUint64;
		result.nUint64 |= temp.nUint64;

	}	//else if -- 64-bit

	return result.nUint64;
}

//KERNEL_VERSION(2,4,0)
int OsaStrLen(const char *s)
{
	int i;

	i = 0;
	while (s[i] != '\0')
		++i;

	return i;

}


void OsaStrCpy(char *s, char *t)
{
	while(( *s = *t) != '\0')
	{
		s++;
		t++;
	}
}


int OsaStrCmp(const char *s, const char *t)
{
	for( ; *s == *t; s++, t++)
		if( *s == '\0')
			return 0;

	return *s - *t;
}

