/* $Workfile: otxdrv_cs.h $   $Revision: 3 $    $Modtime: 5/24/01 12:01p $  */
#ifndef OTX_DRV_CS_IO_H
#define OTX_DRV_CS_IO_H
/*******************************************************************************

  Odin TeleSystems Inc. 2001

  @doc internal

  OTX Generic Hardware Driver API

  @module

  otxdrv_cs |

  IO Control codes and structs for the interface between the OTX Linux PCMCIA Client
  driver (otxdrv_cs) and the OTX Hardware driver (OtxDrv).

  The OtxDrv module needs to be started with the 'nfnOsaLinuxCsIoControl' in order
  to support OTX PCMCIA cards. The 'nfnOsaLinuxCsIoControl' is the address of the 
  OsaLinuxCsIoControl() function in the otxdrv_cs driver (which can be retrieved by
  grepping /proc/ksyms).

*******************************************************************************/

#define IN
#define OUT


// @devnote Remember to step this version whenever this interface changes
#define OTX_DRV_CS_VERSION             0x0002

/*******************************************************************************
  @devnote
   Misc macros
*******************************************************************************/
#define OTX_DRV_CS_DEV_NAME_LEN 32

/*******************************************************************************
  @devnote
   IOCTL codes.
*******************************************************************************/

#define OTX_DRV_CS_IO_GET_VERSION             0x10001
#define OTX_DRV_CS_IO_GET_IO_BASE             0x10002
#define OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK  0x10003
#define OTX_DRV_CS_IO_GET_DEV_NAME            0x10004
#define OTX_DRV_CS_IO_CLR_CALLBACKS           0x10005
#define OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE     0x10006
#define OTX_DRV_CS_IO_GET_CS_DEVICE           0x10007

/*******************************************************************************
  @devnote
   Type definitions
*******************************************************************************/

// Io Control function that will allow the OTX driver to make IO control calls to 
// the PCMCIA Client driver (otxdrv_cs)
typedef int (*OsaLinuxCsIoControlFnP)(
  void *pDevice, 
  unsigned int nCode, 
  void *sIn, 
  unsigned int cbSin,
  void *sOut, 
  unsigned int cbSout
);

// Interrupt callback function
typedef void (*OsaPcmciaBoardNotificationFnP)(
  int nIrq,
  void *pContext,  // pExternalDevice
  struct pt_regs *pRegs
);

// Add Device Callback function. 
// The PCMCIA Client driver (otxdrv_cs) calls this function when an OTX PCMCIA
// card is inserted to notify the OTX driver that this event occurred.
typedef int (*OsaLinuxAddDeviceCallbackFnP)(
  IN void *pCompanionDevice   // (dev_link_t *), referred to as link in the PCMCIA client driver
);


// Remove Device Callback function. 
// The PCMCIA Client driver (otxdrv_cs) calls this function when an OTX PCMCIA
// card is removed to notify the OTX driver that this event occurred.
typedef int (*OsaLinuxRemoveDeviceCallbackFnP)(
  IN void *pExternalDevice
);


#ifdef __cplusplus
extern "C" {
#endif

int OsaLinuxAddDeviceCallback(
  IN void *pCompanionDevice
);

int OsaLinuxRemoveDeviceCallback(
  IN void *pExternalDevice
);


#ifdef __cplusplus
}
#endif

/*******************************************************************************
  @devnote
   Structs for passing data over the IOCTL interface
*******************************************************************************/

typedef struct OsaLinuxCsIoOutGetVersionS
{
  unsigned int        nVersion;
} OsaLinuxCsIoOutGetVersionS; 

typedef struct OsaLinuxCsIoOutGetDevNameS
{
  char szDevName[OTX_DRV_CS_DEV_NAME_LEN];
} OsaLinuxCsIoOutGetDevNameS;

typedef struct OsaLinuxCsIoInSetExternalDeviceS
{
  void *pExternalDevice;
} OsaLinuxCsIoInSetExternalDeviceS;

typedef struct OsaLinuxCsIoOutGetCsDeviceS
{
  void *pCsDevice;
} OsaLinuxCsIoOutGetCsDeviceS;

typedef struct OsaLinuxCsIoOutGetIoBaseS
{
	int nSlot;
	unsigned int nIobase1;
	int nSize1;
	unsigned int nIobase2;
	int nSize2;
} OsaLinuxCsIoOutGetIoBaseS;

typedef struct OsaLinuxCsIoInSetInterruptCallbackS
{
  OsaPcmciaBoardNotificationFnP pfnBoardNotificationCallback;
} OsaLinuxCsIoInSetInterruptCallbackS;

typedef struct OsaLinuxCsIoInSetCallbacksS
{
  OsaLinuxAddDeviceCallbackFnP pfnAddDeviceCallback;
  OsaLinuxRemoveDeviceCallbackFnP pfnRemoveDeviceCallback;
} OsaLinuxCsIoInSetCallbacksS;


/* $History: otxdrv_cs.h $
 * 
 * *****************  Version 3  *****************
 * User: Hanz         Date: 5/25/01    Time: 2:18p
 * Updated in $/Products/Otx-Driver/Osa/Linux/pcmcia/modules
 * Added support for multiple OTX PCMCIA cards in the Linux driver.
 * 
 * *****************  Version 2  *****************
 * User: Hanz         Date: 5/18/01    Time: 1:32p
 * Updated in $/Products/Otx-Driver/Osa/Linux/pcmcia/modules
 * Added OTX_DRV_CS_IO_CLR_CALLBACKS IO call from
 * OsaLinuxReturnResources(). This should prevent the OTX Client driver
 * from calling our interrupt callback once the OTX driver is unloaded.
 * Also we don't try to release the interrupt for the PCMCIA cards (since
 * this is the task of the PCMCIA Client driver). 
 * 
 * *****************  Version 1  *****************
 * User: Hanz         Date: 5/14/01    Time: 10:27a
 * Created in $/Products/Otx-Driver/Osa/Linux/pcmcia/modules
 * Interface header file for the OTX PCMCIA Client driver and the OTX
 * Hardwared driver (i.e. shared by both drivers).
 * 
 * *****************  Version 1  *****************
 * User: Hanz         Date: 5/11/01    Time: 10:09a
 * Created in $/Products/Otx-Driver/Osa/Linux/pcmcia/src/clients
 *
 */

#endif  // OTX_DRV_CS_IO_H
